"""Conversion routines to import tracks."""

from typing import Any, List, Tuple

import gpxpy


class InvalidFileError(Exception):
    """Raised if the file cannot be read correctly."""

    pass


def extract_gpx(track_xml: str, max_hdop: int = 16) -> Tuple[List[Any], List[Any]]:
    """Convert GPX XML string to a list of trackpoints."""
    trackpoints: List[Tuple] = []
    waypoints: List[Tuple] = []
    try:
        gpx = gpxpy.parse(track_xml)
    except gpxpy.gpx.GPXXMLSyntaxException as exc:
        raise InvalidFileError from exc
    for track in gpx.tracks:
        for segment in track.segments:
            for point in segment.points:
                if point.horizontal_dilution <= max_hdop:
                    trackpoints.append((point.time, point.latitude, point.longitude))
    for num, waypoint in enumerate(gpx.waypoints, 1):
        name = waypoint.name or str(num)
        waypoints.append((name, waypoint.latitude, waypoint.longitude))
    return trackpoints, waypoints



__version__ = "0.2.1"

__title__ = "pygohome"
__description__ = "A 100% personal route optimizer in a known environment based on experience."
__url__ = "https://github.com/eumiro/pygohome/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Miroslav Šedivý"
__email__ = "eumiro@gmail.com"

__license__ = "BSD-3"
__copyright__ = "Copyright (c) 2020 Miroslav Šedivý"



"""
Your world.

It consists of tracks, points of interest, intersections,
and a graph that will tell you how to get from A to B.
"""

import datetime as dt
import math
from typing import Dict, List, Tuple

import networkx as nx
import numpy as np

from pygohome.convert import extract_gpx
from pygohome.processor import (
    build_graph,
    find_encounters,
    prepare_trackpoints,
    prepare_waypoints,
    RegionTooLargeError,
)


class World:
    """Your world."""

    trackpoints: List[Tuple[dt.datetime, float, float]]
    waypoints: List[Tuple[str, float, float]]
    graph: nx.DiGraph

    def __init__(self) -> None:
        """Init an empty world."""
        self.trackpoints = []
        self.waypoints = []
        self.graph = None

    def add_trackpoints(self, trackpoints: List) -> None:
        """Add a list of trackpoints."""
        self.trackpoints.extend(trackpoints)
        self.graph = None

    def add_waypoints(self, waypoints: List) -> None:
        """Add a list of waypoints."""
        self.waypoints.extend(waypoints)
        self.graph = None

    def add_gpx(self, track_xml: str) -> None:
        """Add a GPX XML file content."""
        trackpoints, waypoints = extract_gpx(track_xml)
        if trackpoints:
            self.add_trackpoints(trackpoints)
        if waypoints:
            self.add_waypoints(waypoints)

    def _ensure_graph(self) -> None:
        """Rebuild graph if needed."""
        if self.graph is None:
            dfr_trackpoints = prepare_trackpoints(self.trackpoints)
            dfr_waypoints = prepare_waypoints(self.waypoints)
            if set(dfr_trackpoints["utm_zone"]) != set(dfr_waypoints["utm_zone"]):
                raise RegionTooLargeError(
                    f"Trackpoints ({dfr_trackpoints['utm_zone']!r}) and "
                    f"waypoints ({dfr_waypoints['utm_zone']!r}) "
                    f"in different UTM_zones."
                )
            dfr_encounters = find_encounters(dfr_trackpoints, dfr_waypoints)
            self.graph = build_graph(dfr_encounters, dfr_waypoints)

    def fastest_path(self, src: str, dst: str, quantile: float = 0.8) -> nx.Graph:
        """Find the shortest path between src and dst with quantile probability."""
        self._ensure_graph()
        path = nx.path_graph(
            nx.dijkstra_path(
                self.graph, src, dst, lambda u, v, a: np.quantile(a["secs"], quantile)
            )
        )
        return path

    def single_source_periods(self, src: str, quantile: float = 0.8) -> Dict:
        """Return periods to every other waypoint from the src."""
        self._ensure_graph()
        all_dsts_periods = nx.single_source_dijkstra(
            self.graph,
            src,
            weight=lambda u, v, a: int(np.quantile(a["secs"], quantile)),
        )[0]
        periods: Dict = {}
        for dst, period in all_dsts_periods.items():
            if isinstance(dst, tuple):
                # if dst is a tuple(here, src, dst), it is at a lights intersection
                # we want only to get to any node within this intersection
                # result: the shortest period to get to any node near here
                periods[dst[0]] = min(period, periods.get(dst[0], math.inf))
            else:
                periods[dst] = period
        return periods



"""Code from Jupyter notebook.

Call:

    from pygohome.jupygohome import gohome
    gohome()
"""

from typing import Any  # pragma: no cover

import ipyleaflet as lf  # pragma: no cover
import ipywidgets as wd  # pragma: no cover
import numpy as np  # pragma: no cover

from pygohome.world import World  # pragma: no cover


def gohome() -> lf.Map:  # pragma: no cover
    """Execute in a Jupyter Notebook cell."""
    world = World()
    m = lf.Map(
        zoom=14,
        center=(49.0, 8.4),
        interpolation="nearest",
        basemap=lf.basemaps.CartoDB.DarkMatter,
    )

    def action_load(change: Any) -> None:
        uploads = sorted(load_list.value.items())
        load_progress.max = len(uploads)
        for item_i, (_, content) in enumerate(uploads, 1):
            if content["content"]:
                world.add_gpx(content["content"].decode("utf-8"))
            load_progress.value = item_i + 1
        world._ensure_graph()
        pois = sorted(
            node
            for node in world.graph.nodes
            if isinstance(node, str) and not node.isdigit()
        )
        route_src.options = pois
        route_dst.options = pois

    load_list = wd.FileUpload(accept=".gpx", multiple=True)
    load_btn = wd.Button(description="Load GPX files")
    load_btn.on_click(action_load)
    load_progress = wd.IntProgress(
        value=0, min=0, max=0, step=1, bar_style="", orientation="horizontal",
    )

    load_box = wd.VBox(children=[load_list, load_btn, load_progress])

    route = lf.AntPath(locations=[])
    m.add_layer(route)

    route_src = wd.Dropdown(options=[], description="Src:")
    route_dst = wd.Dropdown(options=[], description="Dst:")
    route_slider = wd.FloatSlider(
        value=0.8,
        min=0,
        max=1.0,
        step=0.05,
        description="Quantile:",
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format=".2f",
    )
    route_btn = wd.Button(description="Find route")
    route_result = wd.Text()

    def action_route(change: Any) -> None:
        fp = world.fastest_path(
            route_src.value, route_dst.value, quantile=route_slider.value
        )
        nodes = world.graph.nodes
        route.locations = [
            (nodes[node]["latitude"], nodes[node]["longitude"]) for node in fp.nodes
        ]
        m.center = (
            nodes[route_src.value]["latitude"],
            nodes[route_src.value]["longitude"],
        )
        period_exp, period_min, period_max = 0, 0, 0
        for edge in fp.edges:
            secs = world.graph.edges[edge]["secs"]
            period_exp += np.quantile(secs, route_slider.value)
            period_min += min(secs)
            period_max += max(secs)

        route_result.value = "{}:{:02d} (min: {}:{:02d}, max: {}:{:02d})".format(
            *divmod(int(period_exp), 60),
            *divmod(int(period_min), 60),
            *divmod(int(period_max), 60)
        )

    route_btn.on_click(action_route)
    route_box = wd.VBox(
        children=[route_src, route_dst, route_slider, route_btn, route_result]
    )

    tab_box = wd.Tab(children=[load_box, route_box])
    tab_box.set_title(0, "Load GPX files")
    tab_box.set_title(1, "Fastest route")

    ctrl_tab = lf.WidgetControl(widget=tab_box, position="topright")
    m.add_control(ctrl_tab)
    m.add_control(lf.FullScreenControl())

    return m



"""
pygohome setup.py
"""
import re
from pathlib import Path
from setuptools import setup, find_packages

NAME = "pygohome"
KEYWORDS = ["gps", "navigation", "routing", "bicycle", "walk", "optimal route"]
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Web Environment",
    "Framework :: Jupyter",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    'Operating System :: OS Independent',
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering :: GIS",
]
INSTALL_REQUIRES = [
    "gpxpy >= 1.4.2",
    "ipyleaflet >= 0.13.3",
    "ipywidgets >= 7.5.1",
    "networkx >= 2.5",
    "notebook >= 6.1.5",
    "numpy >= 1.19.4",
    "pandas >= 1.1.5",
    "scipy >= 1.5.4",
    "utm >= 0.7.0",
]

# --+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----


if __name__ == "__main__":
    HERE = Path(__file__).resolve().parent
    META_FILE = (HERE / 'src' / NAME / '__init__.py').read_text()
    META = dict(
        re.findall(r"^__(\w+)__ = ['\"]([^'\"]*)['\"]", META_FILE, re.M))
    setup(
        name=NAME,
        description=META["description"],
        license=META["license"],
        url=META["url"],
        version=META["version"],
        author=META["author"],
        author_email=META["email"],
        maintainer=META["author"],
        maintainer_email=META["email"],
        keywords=KEYWORDS,
        long_description=Path("README.rst").read_text(),
        long_description_content_type="text/x-rst",
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        extras_require={'test': ['pytest']},
        options={},
        include_package_data=True,
    )



"""The processor of pygohome."""

import datetime as dt
from typing import Dict, List, Tuple

import networkx as nx
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
import utm


class ProcessError(Exception):
    """Process Error."""

    pass


class RegionTooLargeError(ProcessError):
    """Region too large."""

    pass


class EmptyDataError(ProcessError):
    """Empty data error."""

    pass


def prepare_trackpoints(trackpoints: List[Tuple[dt.datetime, float, float]]) -> Dict:
    """Prepare the trackpoints to DataFrame."""
    # create 3 Series sorted by timestamp
    if not trackpoints:
        raise EmptyDataError("Trackpoints are empty.")
    raw_dfr = pd.DataFrame.from_records(
        trackpoints, columns=("timestamp", "latitude", "longitude")
    ).sort_values("timestamp")
    timestamp = raw_dfr.pop("timestamp")

    # convert lat/lon to UTM
    dfr = latlon_to_utm(raw_dfr)

    # split into numbered segments (at least 1 minute break between points)
    dfr["segment"] = (timestamp.diff() > pd.Timedelta("00:01:00")).cumsum()

    # calculate offset to the first record of each segment in seconds
    dfr["offset"] = (
        (timestamp - timestamp.groupby(dfr["segment"]).transform("first"))
        .dt.total_seconds()
        .astype(int)
    )

    return dfr


def prepare_waypoints(waypoints: List[Tuple[str, float, float]]) -> Dict:
    """Prepare the waypoints to DataFrame."""
    # create 3 Series sorted by timestamp
    if not waypoints:
        raise EmptyDataError("Waypoints are empty.")
    dfr = pd.DataFrame.from_records(
        waypoints, columns=("name", "latitude", "longitude")
    ).set_index("name")
    return pd.concat([dfr, latlon_to_utm(dfr)], axis=1)


def latlon_to_utm(latlon: pd.DataFrame) -> pd.DataFrame:
    """Convert lat/lon to UTM."""
    utm_x, utm_y, utm_zone, utm_ch = utm.from_latlon(
        latlon["latitude"].values, latlon["longitude"].values
    )
    dfr = pd.DataFrame(
        {
            "utm_x": utm_x.astype(int),
            "utm_y": utm_y.astype(int),
            "utm_zone": utm_zone,
            "utm_ch": utm_ch,
        },
        index=latlon.index,
    )

    # if the region does not fit into a single UTM zone, raise an exception
    if 400000 < np.abs(dfr["utm_x"] - 500000).max():
        raise RegionTooLargeError(
            f"Region too large, does not fit into a single UTM zone: "
            f"lat {latlon['latitude'].min():.2f}…{latlon['latitude'].max():.2f}, "
            f"lon {latlon['longitude'].min():.2f}…{latlon['latitude'].max():.2f}."
        )
    return dfr


def find_encounters(
    dfr_trackpoints: pd.DataFrame, dfr_waypoints: pd.DataFrame, max_dist: int = 30
) -> pd.DataFrame:
    """Find encounters of tracks near waypoints."""
    # Build a KDTree of all nodes and check if trackpoints are closer than 30 meters
    kdtree = cKDTree(dfr_waypoints[["utm_x", "utm_y"]])
    nodes = kdtree.query(
        dfr_trackpoints[["utm_x", "utm_y"]], distance_upper_bound=max_dist
    )[1]
    dfr = pd.DataFrame(
        {
            "segment": dfr_trackpoints["segment"],
            "offset": dfr_trackpoints["offset"],
            "node_iloc": nodes,
        }
    )[nodes < kdtree.n]
    dfr["node"] = dfr_waypoints.index[dfr.node_iloc]

    # for each encounter get the offset of the first and the last point
    dfr_encounters = dfr.groupby(
        (dfr.node != dfr.groupby("segment")["node"].shift(1)).cumsum()
    ).agg(
        segment=("segment", "first"),
        start=("offset", "first"),
        end=("offset", "last"),
        node=("node", "first"),
    )
    return dfr_encounters


def build_graph(
    dfr_encounters: pd.DataFrame, dfr_waypoints: pd.DataFrame
) -> nx.DiGraph:
    """Build a graph with complete information about the route network."""
    # `pred_secs` between leaving `pred_node` and entering `curr_node`
    # `curr_secs` between entering and leaving `curr_node`
    # `succ_secs` between leaving `curr_node` and entering `succ_node`
    dfr_pred = dfr_encounters.groupby("segment").shift(1, fill_value=-1)
    dfr_succ = dfr_encounters.groupby("segment").shift(-1, fill_value=-1)
    dfr = pd.DataFrame(
        {
            "pred_node": dfr_pred["node"],
            "curr_node": dfr_encounters["node"],
            "succ_node": dfr_succ["node"],
            "pred_secs": dfr_encounters["start"] - dfr_pred["end"],
            "curr_secs": dfr_encounters["end"] - dfr_encounters["start"],
            "succ_secs": dfr_succ["start"] - dfr_encounters["end"],
        }
    )

    is_poi = ~dfr.join(dfr_waypoints, on="curr_node")["curr_node"].str.isdigit()

    # an intersection is "slow" (traffic lights)
    # if at least 25% of tracks spend more than 20 seconds within 30 meters of its node
    is_slow = dfr.groupby("curr_node")["curr_secs"].transform(
        lambda x: x.quantile(0.75) > 20
    )

    grp_slow = (
        dfr[~is_poi & is_slow]
        .query("pred_node != -1 and succ_node != -1")
        .groupby(["pred_node", "curr_node", "succ_node"])
    )
    dfr["succ_secs"] += (
        dfr[~is_poi & ~is_slow]["curr_secs"].reindex(dfr.index).fillna(0)
    )
    dfr.loc[~is_poi & ~is_slow, "curr_secs"] = 0

    grp_simple = dfr.query("succ_node != -1").groupby(["curr_node", "succ_node"])

    # build the graph
    graph = nx.DiGraph()
    graph.add_nodes_from(dfr_waypoints.to_dict("index").items())
    graph.add_edges_from(
        (
            ((curr, pred, curr), (curr, curr, succ), {"secs": sorted(grp["curr_secs"])})
            for (pred, curr, succ), grp in grp_slow
        )
    )
    graph.add_edges_from(
        (
            (
                (curr, curr, succ) if (curr, curr, succ) in graph else curr,
                (succ, curr, succ) if (succ, curr, succ) in graph else succ,
                {"secs": sorted(grp["succ_secs"])},
            )
            for (curr, succ), grp in grp_simple
        )
    )
    for node in graph.nodes:
        if isinstance(node, tuple):
            here, src, dst = node
            graph.add_node(node, **graph.nodes[src if here == src else dst])
    return graph
